/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is TotalMessage.
 *
 * The Initial Developer of the Original Code is
 *   alta88 <alta88@gmail.com>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2014
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//*****************************************************************************/
// TotalMessage Preferences functions.

TM.Preferences = {

  get tmPanel()          { return TM.u.d("totalMessagePanel"); },
  get tmPanelBox()       { return TM.u.d("headersPanelBox"); },
  get headersTree()      { return TM.u.d("headersTree"); },
  get headerMapListBox() { return TM.u.d("headerMapListBox"); },

  get messageType() {
    let formatCol = TM.u.d("formatCol");
    let wrapCol = TM.u.d("wrapCol");
    let ellipsisCol = TM.u.d("ellipsisCol");
    let menulist = TM.u.d("headerMessageTypeMenuList");
    if (this.headerType == TM.v.HEADER[TM.v.HEADER.COMPOSE]) {
      wrapCol.disabled = ellipsisCol.disabled = menulist.disabled = true;
      formatCol.disabled = false;
      formatCol.setAttribute("label", formatCol.getAttribute("labelcompose"));
      formatCol.setAttribute("tooltiptext", formatCol.getAttribute("tooltiptextcompose"));
//TM.u.log.debug("TM.Preferences.messageType: compose - "+menulist.selectedItem.value);
      return TM.v.HEADER.MSG_TYPE_MAIL;
    }

    if (this.headerType == TM.v.HEADER[TM.v.HEADER.LIST] ||
        this.headerType == TM.v.HEADER[TM.v.HEADER.LISTMR])
      formatCol.disabled = wrapCol.disabled = ellipsisCol.disabled = true;
    else {
      formatCol.disabled = wrapCol.disabled = ellipsisCol.disabled = false;
    }

    menulist.disabled = false;
    formatCol.setAttribute("label", formatCol.getAttribute("labeldefault"));
    formatCol.setAttribute("tooltiptext", formatCol.getAttribute("tooltiptextdefault"));
//TM.u.log.debug("TM.Preferences.messageType: "+menulist.selectedItem.value);
    return menulist.selectedItem.value;
  },

  get headerType() {
//TM.u.log.debug("TM.Preferences.headerType: "+TM.u.d("headerTypeMenuList").selectedItem.value);
    return TM.u.d("headerTypeMenuList").selectedItem.value;
  },

  get headerMap() {
    return this._headerMap ? this._headerMap :
           this._headerMap = TM.u.mainWindow.TM.u.getHeaderMap(this.messageType);
  },

  get headerMapDefault() {
    if (this._headerMapDefault)
      return this._headerMapDefault;
    let defaultMapName = "headerMap" + this.messageType + "Default";
    if (TM.v[defaultMapName])
      return this._headerMapDefault = TM.v[defaultMapName];
    return this._headerMapDefault = TM.v["headerMapMailDefault"];
  },

  onLoad: function(aEvent) {
TM.u.log.debug("TM.Preferences.onLoad: START ");
    let p = document.getElementById("paneDisplay");
    if (p.parentNode.lastSelected == "paneDisplay")
      TM.Preferences.onPaneLoad(aEvent);
    else
      p.addEventListener("paneload", TM.Preferences.onPaneLoad);
  },

  onPaneLoad: function() {
    setTimeout(function() {
TM.u.log.debug("TM.Preferences.onPaneLoad: START ");
      // The overlay of our large widget creates a timing issue as fonts/tags
      // panels are not initialized properly. So make sure the base tab/tabpanel
      // is set up, gDisplayPane.init() is run, then append our widget and init.
      TM.Preferences.tmPanel.appendChild(TM.Preferences.tmPanelBox).hidden = false;

      TM.Preferences.init();
      TM.Preferences.createMap();
      // Restore open rows.
      TM.Preferences.restoreOpenState();
      TM.Preferences.reflow();
    }, 100);
  },

  onSelectBox: function(aEvent) {
    // Select in headers tree or listbox, or listbox header button oncommand.
TM.u.log.debug("TM.Preferences.onSelectBox: START ");
    TM.Preferences.updateButtonState();
  },

  onSelectTree: function(aEvent) {
    // Select in headers tree or listbox, or listbox header button oncommand.
TM.u.log.debug("TM.Preferences.onSelectTree: START ");
    TM.Preferences.updateButtonState();
  },

  onClick: function(aEvent) {
    // Click on treechildren.
    if (aEvent.button != 0)
      return;
TM.u.log.debug("TM.Preferences.onClick: START ");
    let row = { }, col = { }, element = { }, val;
    this.treeBox.getCellAt(aEvent.clientX, aEvent.clientY, row, col, element);
    let colId = col.value.id;
    if (row.value == -1 || element.value == "twisty" ||
        (colId != "wrapCol" && colId != "ellipsisCol"))
      return;

TM.u.log.debug("TM.Preferences.onClick: element "+element.value+":"+row.value+":"+col.value.id);

    let headerIndex = colId == "wrapCol" ? TM.v.kWrapLines : TM.v.kEllipsis;
    let headerName = this.visibleArray[row.value][this.kName];
    let [headerCategory, headerValue] = this.getHeaderValue(headerName);

    if (headerValue) {
      if (!headerValue[TM.v.kMapped])
        // Don't edit unmapped headers unless it's composeHeader.
        return;

      if (headerIndex == TM.v.kWrapLines)
        headerValue[TM.v.kWrapLines] = headerValue[TM.v.kWrapLines] ? 0 : 1;
      else
        headerValue[headerIndex] = !headerValue[headerIndex];
//      this[headerCategory].set(headerName, headerValue);
      this.treeBox.invalidateRow(row.value);
    }
    else
      TM.u.log.warn("TM.Preferences.onClick: header not found - "+headerName);
  },

  init: function() {
TM.u.log.debug("TM.Preferences.init: START ");
    for (let id of ["headerTypeMenuList", "headerMessageTypeMenuList"]) {
      let ml = TM.u.d(id);
      let selector = "menuitem[value='" + ml.getAttribute("value") + "']";
      ml.selectedItem = ml.querySelectorAll(selector)[0];
TM.u.log.debug("TM.Preferences.init: "+id+" - "+ml.selectedItem.value);
    }

    // Fill in prefs.
    TM.u.d("separatorChar").value = TM.u.HEADER_SEP;
    TM.u.d("wrapLines").value     = TM.u.HEADER_WRAP_LINES;
    TM.u.d("autoCollect").checked = TM.u.AUTOCOLLECT_ACTIVE;
    // Set up strings.
    this["commonHeadersStr"]      = TM.u.strings.get("commonHeaders");
    this["collectedHeadersStr"]   = TM.u.strings.get("collectedHeaders");
    this["customHeadersStr"]      = TM.u.strings.get("customHeaders");
    this["otherValuesStr"]        = TM.u.strings.get("otherValues");
    this["textStr"]               = TM.u.strings.get("text");
    this["emailStr"]              = TM.u.strings.get("email");
    this["messageidStr"]          = TM.u.strings.get("messageid");
    this["newsgroupStr"]          = TM.u.strings.get("newsgroup");
    this["linkStr"]               = TM.u.strings.get("link");
    this["editDiplayNameKey"]     = TM.u.strings.get("editDiplayNameKey");
    this["editFormatKey"]         = TM.u.strings.get("editFormatKey");
    this["toggleWrapKey"]         = TM.u.strings.get("toggleWrapKey");
    this["toggleEllipsisKey"]     = TM.u.strings.get("toggleEllipsisKey");
//    this["toggleComposeColKey"] = "c"; //TM.u.strings.get("toggleComposeColKey");

    // Set the headersTree view.
    this.headersTree.view = this;
TM.u.log.debug("TM.Preferences.init: rowcount - "+ this.headersTree.view.rowCount);
  },

  reflow: function() {
TM.u.log.debug("TM.Preferences.reflow: START ");
    setTimeout(function() {
                 window.opener.TM.Headers.setToolbarState();
                 window.opener.TM.Headers.setWrapState();
               }, 100);
  },

  restoreOpenState: function() {
TM.u.log.debug("TM.Preferences.restoreOpenState: START ");
    let openbits = parseInt(this.headersTree.getAttribute("openbits"));
    for (let row of this.visibleArray) {
      if (row[this.kIsContainer] && openbits & row[this.kOpenBit])
        this.toggleOpenState(this.getRowForCategory(row[this.kCategoryName]));
    }
  },

  setOpenState: function() {
TM.u.log.debug("TM.Preferences.setOpenState: START ");
    let openbits = 0;
    for (let row of this.visibleArray) {
      if (row[this.kIsOpen])
        openbits += row[this.kOpenBit];
    }
    this.headersTree.setAttribute("openbits", openbits);
TM.u.log.debug("TM.Preferences.setOpenState: set openbits - "+openbits);
  },

  updateButtonState: function() {
TM.u.log.debug("TM.Preferences.updateButtonState: START ");
    let listbox = this.headerMapListBox;
    let listboxSelected = listbox.selectedCount > 0;
    let headerstreeSelected = this.headersTree.view &&
                              this.headersTree.view.selection.count > 0;

    if (listboxSelected && listbox.hasAttribute("hasemptyrow") &&
        !listbox.selectedItem.hasAttribute("emptyrow")) {
      // Delete empty non selected row, should only ever be one at this point.
      this.removeRow(listbox.querySelector("listitem[emptyrow]"));
      listbox.removeAttribute("hasemptyrow");
    }

    if (listboxSelected) {
      if (this.headerType != TM.v.HEADER[TM.v.HEADER.LIST]) {
        if (listbox.hasAttribute("hasemptyrow"))
          TM.u.d("addRowButton").setAttribute("disabled", true);
        else
          TM.u.d("addRowButton").removeAttribute("disabled");
      }
      else {
        TM.u.d("addRowButton").setAttribute("disabled", true);
      }
      TM.u.d("removeRowOrHeaderButton").removeAttribute("disabled");
    }
    else {
      TM.u.d("addRowButton").setAttribute("disabled", true);
      TM.u.d("removeRowOrHeaderButton").setAttribute("disabled", true);
    }

    let treeRow = this.headersTree.currentIndex;
//TM.u.log.debug("TM.Preferences.updateButtonState: treeRow - "+treeRow);
    if (headerstreeSelected && !this.isContainer(treeRow) &&
        this.getParentIndex(treeRow) != this.getRowForCategory("commonHeaders"))
      TM.u.d("deleteHeaderButton").removeAttribute("disabled");
    else
      TM.u.d("deleteHeaderButton").setAttribute("disabled", true);

    let isNotNewsListHeader = false;
    if (treeRow >= 0 &&
        (this.headerType == TM.v.HEADER[TM.v.HEADER.LIST] ||
         this.headerType == TM.v.HEADER[TM.v.HEADER.LISTMR]) &&
        this.messageType == TM.v.HEADER.MSG_TYPE_NEWS) {
      // No headers for remote nntp except from msgHdr/cols.
      let headerName = this.visibleArray[treeRow][this.kName];
      let colId = window.opener.TM.u.commonHeaders.has(headerName) ?
                    window.opener.TM.u.commonHeaders.get(headerName)[TM.v.kColId] :
                  window.opener.TM.u.otherValues.has(headerName) ?
                    window.opener.TM.u.otherValues.get(headerName)[TM.v.kColId] : null;
      isNotNewsListHeader = colId ? false : true;
    }

    if (headerstreeSelected && !this.isContainer(treeRow) && !isNotNewsListHeader)
      TM.u.d("addHeaderToMapButton").removeAttribute("disabled");
    else
      TM.u.d("addHeaderToMapButton").setAttribute("disabled", true);
  },


/* Headers map listbox functions **********************************************/

  createMap: function() {
TM.u.log.debug("TM.Preferences.createMap: START Map="+this.headerMap.toSource());
    // Clear old list.
    this.headerMapListBox.removeAttribute("hasemptyrow");
    for (let listitem of this.headerMapListBox.querySelectorAll("listitem"))
      listitem.remove();

    if (!(this.headerType in this.headerMap))
      // For v2 headerType additions not in current user set maps.
      this.headerMap[this.headerType] = this.headerMapDefault[this.headerType];

    // Create categories.
    this.categories.map(headerCategory => this.getHeadersCategory(headerCategory));

    for (let [row, headerRow] in Iterator(this.headerMap[this.headerType])) {
TM.u.log.debug("TM.Preferences.createMap: type:row#:rowValue - "+
               this.headerType+":"+row+":"+headerRow.toSource());
      let headerListitem = this.createBlankListitem();

      for (let [headerName, headerValue] in Iterator(headerRow)) {
TM.u.log.trace("TM.Preferences.createMap: headerName - "+headerName);
        if (this.headerExists(headerName))
          // A header in the map may have been deleted from the headers list,
          // don't add it to the listbox.
          headerListitem.firstChild.firstChild
                        .appendChild(this.createHeaderContainer(headerName));
      }

      if (headerListitem.firstChild.firstChild.hasChildNodes())
        this.headerMapListBox.appendChild(headerListitem);
    }

    // Force listbox horizontal scroll, ridiculous hoop jumping!!
    this.fixIdioticForcedScrollHidden();
  },

  createHeaderContainer: function(aName) {
TM.u.log.debug("TM.Preferences.createHeaderContainer: START "+aName);
    let headerContainer = document.createElementNS(TM.v.XUL_NS, "button");
    headerContainer.classList.add("header-button");
    headerContainer.setAttribute("label", aName);
    headerContainer.setAttribute("type", "checkbox");
    return headerContainer;
  },

  addHeader: function(aHeaderName) {
TM.u.log.debug("TM.Preferences.addHeader: START headername - "+aHeaderName); //+":"+this.rowCount);
    let headerName = aHeaderName ? aHeaderName.trim() : "";

    let [headerCategory, headerValue] = this.getHeaderValue(headerName);
    if (headerValue)
      // Already exists.
      return;
//TM.u.log.debug("TM.Preferences.addHeader: rowcount1 - "+ TM.Preferences.headersTree.view.rowCount);
TM.u.log.debug("TM.Preferences.addHeader: rowcount2 - "+ this.headersTree.view.rowCount);

    headerCategory = "customHeaders";
    let customHeaderRow = this.getRowForCategory(headerCategory);
    let newHeader = [].concat(this.headerValueTemplate);
    let updateHeadersSet = TM.u[headerCategory];
    updateHeadersSet.add(headerName);
    TM.u[headerCategory] = updateHeadersSet;
    this[headerCategory].set(headerName, newHeader);
TM.u.log.debug("TM.Preferences.addHeader: row:visarray - "+
               customHeaderRow+":"+this.visibleArray[customHeaderRow]);

    this.headersTree.view.selection.selectEventsSuppressed = true;
    this.treeBox.ensureRowIsVisible(customHeaderRow);
    if (this.visibleArray[customHeaderRow][this.kIsOpen])
      // If folder open, close it.  An open builds the new list.
      this.toggleOpenState(customHeaderRow);
    this.toggleOpenState(customHeaderRow);
//    this.headersTree.view.selection.selectEventsSuppressed = false;
//    this.treeBox.invalidate();
    let row = this.headersTree.view.rowCount - 1;
    let col = this.headersTree.columns.getNamedColumn("nameCol");
    this.treeBox.scrollToRow(row);
    this.treeBox.ensureRowIsVisible(customHeaderRow);
    this.isNameColEditable = false;
    this.headersTree.view.selection.select(row);
    this.headersTree.view.selection.selectEventsSuppressed = false;

    if (!aHeaderName) {
TM.u.log.debug("TM.Preferences.addHeader: START editor ");
      // Start the editor if not passed in the new name of the just edited row.
      this.isNameColEditable = true;
      this.headersTree.startEditing(row, col);
    }
  },

  deleteHeader: function(aHeaderName) {
TM.u.log.debug("TM.Preferences.deleteHeader: START headername - "+aHeaderName);
    if (!aHeaderName && !this.headersTree.view.selection.count)
      return;

    let headerName, row, containerRow;
    let scrollToRow = this.treeBox.getFirstVisibleRow();

    if (aHeaderName)
      headerName = aHeaderName;
    else if (this.isNameColEditable)
      // In custom header editor and clicked somewhere to select a random row
      // with a blank custom headername.
      headerName = aHeaderName;
    else {
      row = this.headersTree.currentIndex;
      headerName = this.visibleArray[row][this.kName];
    }

    let [headerCategory, headerValue] = this.getHeaderValue(headerName);
    let updateHeadersSet = TM.u[headerCategory];
    updateHeadersSet.delete(headerName);
    TM.u[headerCategory] = updateHeadersSet;
    this[headerCategory].delete(headerName);

TM.u.log.debug("TM.Preferences.deleteHeader: headerCategory:headerName:headerValue - "+
               headerCategory+":"+headerName+":"+headerValue);
TM.u.log.trace("TM.Preferences.deleteHeader: headerCategory pref - "+
               JSON.stringify([...TM.u[headerCategory]]));
TM.u.log.trace("TM.Preferences.deleteHeader: headerCategory cache - "+
               JSON.stringify([...this[headerCategory]]));

    // Back to the tree.
    if (this.isNameColEditable) {
      // Deleting our pseudo placeholder row, only in customHeaders folder.
      containerRow = row = this.getRowForCategory("customHeaders");
      row = this.headersTree.view.rowCount - 1;
    }
    else {
      // Deleting selected row, not our pseudo placeholder row.
      containerRow = this.getParentIndex(row);
    }

    this.headersTree.view.selection.selectEventsSuppressed = true;
    this.toggleOpenState(containerRow);
    this.toggleOpenState(containerRow);
    row = row < this.headersTree.view.rowCount ? row : row - 1;
    this.headersTree.view.selection.select(row);
TM.u.log.debug("TM.Preferences.deleteHeader: row - "+row);
//    this.headersTree.currentIndex = row;
    this.treeBox.scrollToRow(scrollToRow);
    this.headersTree.view.selection.selectEventsSuppressed = false;

TM.u.log.debug("TM.Preferences.deleteHeader: currentIndex - "+this.headersTree.currentIndex);

    // Delete header from listbox map.
    let anyMapped = this.headerMapListBox
                        .querySelectorAll('button[label="' + headerName + '"]');
    for (let mapped of anyMapped) {
      let rowNode = mapped.parentNode.parentNode.parentNode;
      mapped.remove();
      if (!rowNode.firstChild.firstChild.hasChildNodes()) {
        // This row is now empty.  Set hasemptyrow to make sure all rows are
        // deleted (unlike deleting multiselection items in the listbox, where
        // one empty row remains).
        rowNode.setAttribute("emptyrow", true);
        this.headerMapListBox.setAttribute("hasemptyrow", true);
        this.handleEmptyRow(rowNode);
      }
    }

    // Finally, delete the header from all maps and save.
    // TODO:
  },

  addHeaderToMap: function() {
TM.u.log.debug("TM.Preferences.addHeaderToMap: START ");
    if ((this.headerType != TM.v.HEADER[TM.v.HEADER.LIST] &&
         !this.headerMapListBox.selectedItem) ||
        (this.headerType == TM.v.HEADER[TM.v.HEADER.LIST] &&
         !this.headerMapListBox.getItemAtIndex(0)) ||
        (this.headerType == TM.v.HEADER[TM.v.HEADER.COMPOSE] &&
         !this.headerMapListBox.hasAttribute("hasemptyrow")))
      this.addBlankListitem();

    let row = this.headersTree.currentIndex;
    let headerName = this.visibleArray[row][this.kName];
    let button = this.createHeaderContainer(headerName);
    if (!this.headerMapListBox.selectedItem)
      this.headerMapListBox.selectItem(this.headerMapListBox.getItemAtIndex(0));
    this.headerMapListBox.selectedItem.firstChild.firstChild.appendChild(button);

    let [headerCategory, headerValue] = this.getHeaderValue(headerName);
    headerValue[TM.v.kMapped] = true;
    // Auto create displayName first time for categories with no defaults;
    // capitalize words per convention.
    if (//this.headerType == TM.v.HEADER[TM.v.HEADER.COMPOSE] &&
        !TM.headerDisplaySpecial(this.messageType, headerName) &&
        headerValue[TM.v.kDisplayName] !== null &&
        headerCategory != "commonHeaders" &&
        headerCategory != "otherValues") {
      let sep = this.headerType != TM.v.HEADER[TM.v.HEADER.COMPOSE] ? TM.u.HEADER_SEP : "";
      let sepL = "", sepR = "";
      TM.u.isRtl ? sepL = sep : sepR = sep;
      let displayName = sepL +
                        headerName.split("-")
                                  .map(h => h[0].toUpperCase() + h.slice(1))
                                  .join("-") +
                        sepR;
      headerValue[TM.v.kDisplayName] = displayName;
    }

TM.u.log.debug("TM.Preferences.addHeaderToMap: B headerCategory:headerName:headerValue - "+
               headerCategory+":"+headerName+":"+headerValue);
//TM.u.log.debug("TM.Preferences.addHeaderToMap: headerCategory cache2 - "+
//               JSON.stringify([...this[headerCategory]]));
    this.treeBox.invalidateRow(row);

    if (this.headerMapListBox.selectedItem.hasAttribute("emptyrow")) {
      this.headerMapListBox.removeAttribute("hasemptyrow");
      this.headerMapListBox.selectedItem.removeAttribute("emptyrow");
    }

    this.updateButtonState();
    TM.u.d("undoButton").removeAttribute("disabled");
  },

  addRow: function() {
TM.u.log.debug("TM.Preferences.addRow: START ");
    this.addBlankListitem();
    this.updateButtonState();
    TM.u.d("undoButton").removeAttribute("disabled");
  },

  addBlankListitem: function() {
TM.u.log.debug("TM.Preferences.addBlankListitem: START ");
    let node = this.headerMapListBox.selectedItem;
    let blankListitem = this.createBlankListitem();
    blankListitem.setAttribute("emptyrow", true);

    if (node && node.tagName == "listitem")
      this.headerMapListBox.insertBefore(blankListitem, node.nextSibling);
    else
      this.headerMapListBox.appendChild(blankListitem);

    this.headerMapListBox.setAttribute("hasemptyrow", true);
    this.headerMapListBox.suppressOnSelect = true;
    this.headerMapListBox.selectItem(blankListitem);
    this.headerMapListBox.suppressOnSelect = false;
    this.headerMapListBox.ensureElementIsVisible(blankListitem);
  },

  createBlankListitem: function() {
TM.u.log.debug("TM.Preferences.createBlankListitem: START ");
    let headerListitem = document.createElementNS(TM.v.XUL_NS, "listitem");
    headerListitem.setAttribute("allowevents", true);
    let headersCell = document.createElementNS(TM.v.XUL_NS, "listcell");
    headersCell.setAttribute("flex", "1");
    let headersCellBox = document.createElementNS(TM.v.XUL_NS, "hbox");
    headersCellBox.classList.add("headersCellHeadersBox");
    headersCell.appendChild(headersCellBox);
    headerListitem.appendChild(headersCell);
    return headerListitem;
  },

  removeRowOrHeader: function() {
TM.u.log.debug("TM.Preferences.removeRowOrHeader: START ");
    let node = this.headerMapListBox.selectedItem;
    if (!node || node.tagName != "listitem")
      return;

    let removedHeaders = [];
    let selectedHeaders = node.querySelectorAll("button[checked]");
    if (!selectedHeaders.length) {
      // Remove all headers if none selected.
      selectedHeaders = node.querySelectorAll("button");
      for (let selectedHeader of selectedHeaders)
        removedHeaders.push(selectedHeader.label);
      this.removeRow(node);
    }
    else {
      for (let selectedHeader of selectedHeaders) {
        removedHeaders.push(selectedHeader.label);
        selectedHeader.remove();
      }

      if (!node.firstChild.firstChild.hasChildNodes())
        // This row is now empty.
        this.handleEmptyRow(node);
    }

    // Set mapped flag to false if a header is completely removed from the
    // map and make the tree reflect the status.
    let gotOne = false;
TM.u.log.debug("TM.Preferences.removeRowOrHeader: removedHeaders "+removedHeaders);
    for (let headerName of removedHeaders) {
      let anyMapped = this.headerMapListBox
                          .querySelectorAll('button[label="' + headerName + '"]');
      if (anyMapped.length == 0) {
        let [headerCategory, headerValue] = this.getHeaderValue(headerName);
        if (headerCategory) {
          // If not already deleted from tree.
          headerValue[TM.v.kMapped] = false;
          headerValue[TM.v.kDisplayName] = "";
          gotOne = true;
        }
      }
    }

    if (gotOne)
      this.treeBox.invalidate();

//    this.updateButtonState();
    TM.u.d("undoButton").removeAttribute("disabled");
  },

  removeRow: function(aNode) {
TM.u.log.debug("TM.Preferences.removeRow: START ");
    let selectNode;
    let node = aNode;
    if (!node || node.tagName != "listitem")
      return;

    if (node.parentNode.currentIndex != 0 ||
        node.nextSibling || !node.selected) {
      // Remove a row if it's not the only one or an unselected empty row.
      if (node.selected && node.hasAttribute("emptyrow"))
        this.headerMapListBox.removeAttribute("hasemptyrow");
      if (node.selected) {
        selectNode = node.parentNode.lastChild == node ? node.previousSibling :
                                                         node.nextSibling;
        node.parentNode.suppressOnSelect = true;
        node.parentNode.selectItem(selectNode);
        node.parentNode.suppressOnSelect = false;
      }
      node.remove();
    }
    else {
      // Remove all of the only row's items.
      let box = node.firstChild.firstChild;
      while (box.lastChild)
        box.lastChild.remove();
      // This is the new empty row.  Update states.
      node.setAttribute("emptyrow", true);
      this.headerMapListBox.setAttribute("hasemptyrow", true);
      this.updateButtonState();
    }
  },

  handleEmptyRow: function(aNode) {
TM.u.log.debug("TM.Preferences.handleEmptyRow: START hasemptyrow - "+
               this.headerMapListBox.hasAttribute("hasemptyrow"));
    if (this.headerMapListBox.hasAttribute("hasemptyrow"))
      // An empty row exists, remove this one.  Can only have 1 empty row.
      this.removeRow(aNode);
    else {
      // This is the new empty row.  Update states.
      this.headerMapListBox.setAttribute("hasemptyrow", true);
      aNode.setAttribute("emptyrow", true);
    }

    this.updateButtonState();
  },

  onRestore: function() {
TM.u.log.debug("TM.Preferences.onRestore: START ");
    // Clear pref for auto rebuild from default.
    let prefName =  "headerMap" + this.messageType + "Pref";
    Services.prefs.clearUserPref(TM.v[prefName]);

    this._headerMap = null;
    this._headerMapDefault = null;
    this.categories.map(headerCategory => this[headerCategory] = null);
TM.u.log.debug("TM.Preferences.onRestore: default map "+this.headerMap.toSource());

    this.createMap();
    this.onApply();
    this.headerMapListBox.removeAttribute("hasemptyrow");
    this.treeBox.invalidate();
  },

  onUndo: function(aWhat) {
TM.u.log.debug("TM.Preferences.onUndo: START ");
    this._headerMap = null;
    this._headerMapDefault = null;
    this.categories.map(headerCategory => this[headerCategory] = null);

    this.createMap();
    this.treeBox.invalidate();
    if (this.headerType == TM.v.HEADER[TM.v.HEADER.LIST] ||
        this.headerType == TM.v.HEADER[TM.v.HEADER.LISTMR])
      TM.u.mainWindow.GetThreadTree().treeBoxObject.invalidate();

    this.updateButtonState();
  },

  onApply: function(aClose) {
    // Create the new map and store it.  Reload headers in the current message.
TM.u.log.debug("TM.Preferences.onApply: START ");
    let rowName, headerName, headerValue;
    let map = this.headerMap;
    let rowDesc = "row";
    let row = 0;
    map[this.headerType] = {};
    for (let listitem of this.headerMapListBox.querySelectorAll("listitem")) {
      let buttons = listitem.querySelectorAll("button");
      if (buttons.length > 0) {
        rowName = rowDesc + (++row);
        map[this.headerType][rowName] = {};
        for (let button of buttons) {
          headerName = button.label;
TM.u.log.trace("TM.Preferences.onApply: row:headerName "+rowName+":"+headerName);
          [, headerValue] = this.getHeaderValue(headerName);
          if (!headerValue)
            continue;
          headerValue.splice(TM.v.kMapped, 1);
          map[this.headerType][rowName][headerName] = headerValue;
        }
      }
    };

    // Last empty row.
    rowName = rowDesc + (++row);
    map[this.headerType][rowName] = {};

    TM.u.d("undoButton").setAttribute("disabled", true);
TM.u.log.debug("TM.Preferences.onApply: new map "+map.toSource());

    // Update prefs.
    TM.u.HEADER_SEP        = TM.u.d("separatorChar").value;
    TM.u.HEADER_WRAP_LINES = TM.u.d("wrapLines").value;
    TM.u.mainWindow.TM.u.setWrapLinesAndBorder();

    // Save the headers map.
    TM.u.mainWindow.TM.u.setHeaderMap(this.messageType, map);
    // Initiate a headers rebuild.
    TM.u.mainWindow.TM.resetHeaderFlags();
    TM.u.mainWindow.TM.Headers.toggle();
    if (!aClose)
      this.onUndo();
  },

  fixIdioticForcedScrollHidden: function() {
TM.u.log.trace("TM.Preferences.fixIdioticForcedScrollHidden: START ");
    let listboxbody, rules, rule;
//    let listitem = this.headerMapListBox.querySelector("listitem");
//    listboxbody = listitem.boxBoject.parentBox;
    let nodes = document.getAnonymousNodes(this.headerMapListBox);
    for (let [, node] in Iterator(nodes)) {
      if (node.localName == "listrows" &&
          node.lastChild.localName == "listboxbody") {
        listboxbody = node.lastChild;
        break;
      }
    }

    if (!listboxbody)
      return;

    // Because this would be just too fricken easy.
    // listboxbody.style.removeProperty("overflow-x");

    let domSvc = Cc["@mozilla.org/inspector/dom-utils;1"]
                   .getService(Ci["inIDOMUtils"]);
    rules = domSvc.getCSSStyleRules(listboxbody);
    for (let j = 0; j < rules.Count(); j++) {
      rule = rules.GetElementAt(j);
      if (rule.selectorText == "listboxbody") {
        rule.QueryInterface(Ci.nsIDOMCSSStyleRule)
            .style.removeProperty("overflow-x");
TM.u.log.debug("TM.Preferences.fixIdioticForcedScrollHidden: fixed - "+rule.cssText);
      }
    }
  },

/* Headers tree functions *****************************************************/

  categories: ["commonHeaders", "collectedHeaders", "customHeaders", "otherValues"],

  // Indices of visibleArray array.
  kName:         0,
  kIsContainer:  1,
  kIsOpen:       2,
  kCategoryName: 3,
  kOpenBit:      4,

  visibleArray: [
    // ["Name String", isContainer, isOpen, "children category map name"]
    ["commonHeadersStr", true, false, "commonHeaders", 1],
    ["collectedHeadersStr", true, false, "collectedHeaders", 2],
    ["otherValuesStr", true, false, "otherValues", 4],
    ["customHeadersStr", true, false, "customHeaders", 8]
  ],

  // [Wraplines, Ellipsis, Format, Display Name Override, isMapped]
  headerValueTemplate: [0, true, "", "", false],

  onKeyPress: function(aEvent) {
    // |which| is either the charCode or the keyCode, depending on which is set.
TM.u.log.debug("TM.Preferences.onKeyPress: START which:charCode:keyCode:key - "+
               aEvent.which+":"+aEvent.charCode+":"+aEvent.keyCode+":"+aEvent.key);
    if (aEvent.altKey || aEvent.metaKey || aEvent.ctrlKey)
      return;

    let headerCategory, headerName, headerValue, col;
    let row = this.headersTree.currentIndex;
    if (this.headersTree.hasAttribute("editing")) {
      headerName = this.visibleArray[this.headersTree._editingRow][this.kName];
      [, headerValue] = this.getHeaderValue(headerName);
      if (this.headersTree._editingColumn &&
          (this.headersTree._editingColumn.id == "nameCol" ||
           (this.headersTree._editingColumn.id == "displayNameCol" &&
            this.headerType == TM.v.HEADER[TM.v.HEADER.COMPOSE]))) {
        // Per RFC5322, header name (email) chars must be printable ascii
        // chars d'33' to d'126' minus the colon d'58'.  So edit check if
        // editing headerName column, and also displayName column if mapping
        // header type 'compose'.
TM.u.log.debug("TM.Preferences.onKeyPress: editing ");
        aEvent.stopPropagation();
        if (!this.hasValidChar(aEvent.key) && !aEvent.keyCode)
          aEvent.preventDefault();
      }

      return;
    }

    if (aEvent.key == this["editDiplayNameKey"]) {
      col = this.headersTree.columns.getNamedColumn("displayNameCol");
      this.headersTree.startEditing(row, col);
      return;
    }
    else if (aEvent.key == this["editFormatKey"]) {
      col = this.headersTree.columns.getNamedColumn("formatCol");
      if (col.disabled || this.getCellProperties(row, col).contains("unMapped"))
        return;
      this.openFormatPopup(row, col);
      this.headersTree.startEditing(row, col);
      return;
    }
    else if (aEvent.key == " " || aEvent.key == KeyEvent.DOM_VK_RIGHT) {
      this.addHeaderToMap();
      return;
    }
    else if (aEvent.key == KeyEvent.DOM_VK_RETURN) {
      aEvent.preventDefault();
      this.addHeaderToMap();
      return;
    }
    else {
      headerName = this.visibleArray[row][this.kName];
      [headerCategory, headerValue] = this.getHeaderValue(headerName);
      if (aEvent.key == this["toggleWrapKey"]) {
        headerValue[TM.v.kWrapLines] = headerValue[TM.v.kWrapLines] ? 0 : 1;
        this.treeBox.invalidateRow(row);
      }
      else if (aEvent.key == this["toggleEllipsisKey"]) {
        headerValue[TM.v.kEllipsis] = !headerValue[TM.v.kEllipsis];
        this.treeBox.invalidateRow(row);
      }
    }
  },

  onInput: function(aEvent) {
    let inputField = aEvent.target._inputField;
    let val = inputField.value;
    let index = val.indexOf(" ");
    if (index == -1)
      return;
TM.u.log.debug("TM.Preferences.onInput: spacehit val:index - "+val+":"+index);
    inputField.value = val.replace(/\s/g, "␠");
    inputField.selectionStart = inputField.selectionEnd = index + 1;
  },

  hasValidChar: function (aStr) {
    function validChar(e) {
      return e.charCodeAt(0) != 58 &&
             e.charCodeAt(0) >= 33 &&
             e.charCodeAt(0) <= 126;
    }

    return Array.every(aStr, validChar);
  },

  getRowForCategory: function(aCategory) {
    let r;
    let lastRow = this.headersTree.view.rowCount - 1;
    for (r = lastRow; r >= 0; r--) {
      if (this.getLevel(r) == 0 &&
          this.visibleArray[r][this.kCategoryName] == aCategory)
        break;
    }

    return r;
  },

  openFormatPopup: function(aRow, aCol) {
    let textX = {}, textY = {};
    // Offsets for positioning the popup.
    let  x = this.treeBox.screenX - window.mozInnerScreenX;
    // Don't forget the titlebar..
    let  y = this.treeBox.screenY - window.screenY;
    this.treeBox.getCoordsForCellItem(aRow+1, aCol, "text", textX, textY, {}, {});
    TM.u.d("headerMessageFormatPopup").openPopup(null, null,
                                                 textX.value + x,
                                                 textY.value + y,
                                                 false, false);
  },

  setFormat: function(aValue) {
TM.u.log.debug("TM.Preferences.setFormat: START "+aValue);
    if (aValue && TM.v.HEADER.FORMAT_TYPES.has(aValue)) {
      let row = this.headersTree.currentIndex;
      let headerName = this.visibleArray[row][this.kName];
      let [, headerValue] = this.getHeaderValue(headerName);
      headerValue[TM.v.kFormat] = aValue;
      this.treeBox.invalidateRow(row);
    }

    this.headersTree.focus();
  },
/*
  handleComposeColSelect: function(aHeaderName) {
TM.u.log.debug("TM.Preferences.handleComposeColSelect: START - "+aHeaderName);
    let headerName = aHeaderName;
    let [headerCategory, headerValue] = this.getHeaderValue(headerName);
    let val = headerValue[TM.v.kComposeHeader];
    this[headerCategory][headerName][TM.v.kComposeHeader] = !val;
    let newComposeMap = this.headerMapCompose;
    if (val) {
      delete newComposeMap[headerName];
      this._headerMapCompose = newComposeMap;
    }
    else {
      let curDisplayName = this.headerMapCompose[headerName] ?
                               this.headerMapCompose[headerName][0] :
                               "";
      let displayName = "";
      if (headerCategory != this.categories[0]) {
        // If not commonHeaders (have strings for those) autofill a name.
        displayName = headerName.toString();
        displayName = displayName.substr(0,1).toUpperCase() +
                      displayName.substr(1);
      }
      newComposeMap[headerName] = curDisplayName ? [curDisplayName] :
                                                   [displayName];
      this._headerMapCompose = newComposeMap;
    }
  },
*/
  getHeadersCategory: function(aCategory) {
//TM.u.log.debug("getHeadersCategory: START "+aCategory);
    if (this[aCategory])
      return;
TM.u.log.debug("TM.Preferences.getHeadersCategory: "+aCategory+":"+
               JSON.stringify([...window.opener.TM.u[aCategory]]));
    let map = new Map();
    // Seed the map with default values.
    for (let headerName of window.opener.TM.u[aCategory].keys()) {
//TM.u.log.debug("TM.Preferences.getHeadersCategory: headerName "+ headerName);
      map.set(headerName, [].concat(this.headerValueTemplate));
    }

    // Apply values from the default map for this message type and header type.
    for (let [rowName, headerRow] in Iterator(this.headerMapDefault[this.headerType])) {
      for (let [headerName, headerValue] in Iterator(headerRow))
        if (map.has(headerName)) {
          headerValue[TM.v.kMapped] = false;
          map.set(headerName, headerValue);
        }
    }

    // Finally apply values from the user configured map for this message type
    // and header type.
    for (let [rowName, headerRow] in Iterator(this.headerMap[this.headerType])) {
      for (let [headerName, headerValue] in Iterator(headerRow))
        if (map.has(headerName)) {
          headerValue[TM.v.kMapped] = true;
          map.set(headerName, headerValue);
        }
    }
TM.u.log.debug("TM.Preferences.getHeadersCategory: map "+
               JSON.stringify([...map]));

    this[aCategory] = map;
  },

  getHeaderValue: function(aHeaderName) {
//TM.u.log.debug("getHeaderValue: START "+aHeaderName);
    // Find the header's values in one of the categories.
    let headerValue, category;
    for (let headerCategory of this.categories) {
      if (this[headerCategory].has(aHeaderName)) {
        category = headerCategory;
        headerValue = this[headerCategory].get(aHeaderName);
        break;
      }
    }

    return [category, headerValue];
  },

  headerExists: function(aHeaderName) {
    // Find the header's values in one of the categories.
    for (let headerCategory of this.categories) {
      if (this[headerCategory].has(aHeaderName))
        return true;
    }

    return false;
  },

/* nsITreeView implementation *************************************************/

  treeBox: null,
  selection: null,
  get rowCount()                         { return this.visibleArray.length; },
  setTree: function(aTreebox)            { this.treeBox = aTreebox; },
  isContainerOpen: function(aRow)        { return this.visibleArray[aRow][this.kIsOpen]; },
  isContainerEmpty: function(aRow)       { return false; },
  isSeparator: function(aRow)            { return false; },
  isSorted: function()                   { return false; },
  isSelectable: function(aRow, aColumn)  { return false; },

  getImageSrc: function(aRow, aCol)      { return null; },
  getProgressMode : function(aRow, aCol) {},
  cycleHeader: function(aCol)            {},
  cycleCell: function(aRow, aCol)        {},
  selectionChanged: function()           {},
  performAction: function(aAction)       {},
  performActionOnCell: function(aAction, aRow, aCol) {},
  getRowProperties: function(aRow)       { return ""; },
  getColumnProperties: function(aCol)    { return ""; },

  isContainer: function(aRow)            {
    return aRow > this.rowCount ? false : this.visibleArray[aRow][this.kIsContainer];
  },

  getCellProperties: function(aRow, aCol) {
    let props = [];
    if (!this.isContainer(aRow)) {
      let headerName = this.visibleArray[aRow][this.kName];
      let [headerCategory, headerValue] = this.getHeaderValue(headerName);
//TM.u.log.debug("getCellProperties: headerCategory:headerName:headerValue - "+
//               headerCategory+":"+headerName+":"+headerValue);

      if (aCol.id == "nameCol")
        props.push("headerNameCol");
      if (aCol.id == "displayNameCol")
        props.push("displayNameCol");

      if (!headerValue ||
          !headerValue[TM.v.kMapped] ||
          aCol.id == "customColumnCol") {
        // Header not in a map, or inapplicable column.
        props.push("unMapped");
      }

      if (this.headerType == TM.v.HEADER[TM.v.HEADER.COMPOSE] &&
          ((aCol.id != "nameCol" && aCol.id != "displayNameCol" &&
            aCol.id != "formatCol") ||
           (aCol.id == "formatCol" &&
            TM.v.composeHeadersDefault.has(headerName)))) {
        // Compose is special.
        props.push("unMapped");
      }

      if (this.headerType != TM.v.HEADER[TM.v.HEADER.COMPOSE] &&
          aCol.id == "formatCol" && headerName == "tags") {
        // No tags format.
        props.push("unMapped");
      }

      if ((this.headerType == TM.v.HEADER[TM.v.HEADER.LIST] ||
           this.headerType == TM.v.HEADER[TM.v.HEADER.LISTMR]) &&
          aCol.id != "nameCol" && aCol.id != "displayNameCol") {
        // List is special.
        props.push("unMapped");
      }
    }

    return props.join(" ");
  },

  getCellText: function(aRow, aCol) {
    if (aRow >= this.visibleArray.length)
      return null;
//TM.u.log.debug("TM.Preferences.getCellText: row - "+aRow);
    let headerName = this.visibleArray[aRow][this.kName];
    if (this.isContainer(aRow)) {
      if (aCol.id == "nameCol")
        // A string.
        return this[headerName];
      else
        return null;
    }

    let headerValueIndex;
    if (aCol.id == "nameCol")
      return headerName;
    if (aCol.id == "displayNameCol")
      headerValueIndex = TM.v.kDisplayName;
    else if (aCol.id == "formatCol")
      headerValueIndex = TM.v.kFormat;
    else
      return null;

    let [headerCategory, headerValue] = this.getHeaderValue(headerName);
    if (!headerValue)
      return null;

    if (headerValueIndex == TM.v.kFormat) {
      if (headerValue[TM.v.kMapped] && headerValue[headerValueIndex]) {
        if (this.headerType == TM.v.HEADER[TM.v.HEADER.COMPOSE])
          // Get the default compose value.
          return headerValue[headerValueIndex].replace(/\s/g, "␠");
        else
          // Get the string for the format token, if exists.
          return this[headerValue[headerValueIndex] + "Str"];
      }
    }

    if (headerValueIndex == TM.v.kDisplayName) {
      if (headerValue[headerValueIndex])
        // A customized value.
        return headerValue[headerValueIndex].replace(/\s/g, "␠");
      if (headerValue[headerValueIndex] === null)
        // User no want.
        return null;

      // Show special case overrides, or dtd names, or nothing.
      let commonHeaderName = window.opener.TM.u["commonHeaders"].get(headerName);
      commonHeaderName = commonHeaderName ? commonHeaderName[TM.v.kLabelValue] : null;
      let otherValuesName = window.opener.TM.u["otherValues"].get(headerName);
      otherValuesName = otherValuesName ? otherValuesName[TM.v.kLabelValue] : null;
      let displayName = TM.headerDisplaySpecial(this.messageType, headerName) ||
                        commonHeaderName || otherValuesName || null;
      let sep = this.headerType != TM.v.HEADER[TM.v.HEADER.COMPOSE] ? TM.u.HEADER_SEP : "";
      let sepL = "", sepR = "";
      TM.u.isRtl ? sepL = sep : sepR = sep;
      return displayName ? sepL + displayName + sepR : null;
    }

    return null;
  },

  getCellValue: function(aRow, aCol) {
    if (this.isContainer(aRow))
      return null;

    let headerValueIndex;
    if (aCol.id == "wrapCol")
      headerValueIndex = TM.v.kWrapLines;
    else if (aCol.id == "ellipsisCol")
      headerValueIndex = TM.v.kEllipsis;
    else
      return null;

    let headerName = this.visibleArray[aRow][this.kName];
    let [headerCategory, headerValue] = this.getHeaderValue(headerName);
    return headerValue ? headerValue[headerValueIndex] ? true : false : null;
  },

  getParentIndex: function(aRow) {
    if (this.isContainer(aRow))
      return -1;
    for (let t = aRow - 1; t >= 0 ; t--) {
      if (this.isContainer(t))
        return t;
    }
    return -1;
  },

  getLevel: function(aRow) {
    if (this.isContainer(aRow))
      return 0;
    return 1;
  },

  isEditable: function(aRow, aCol) {
TM.u.log.debug("TM.Preferences.isEditable: "+aRow);
    if (this.isContainer(aRow))
      return false;

    if (aCol.id == "nameCol" && this.isNameColEditable)
      // Custom header is being added.  The row is always added at the bottom,
      // and is not editable once it's added.
      return true;

    if (aCol.id != "displayNameCol" && aCol.id != "formatCol")
      return false;

    let headerName = this.visibleArray[aRow][this.kName];

    // Autofilled compose custom header values don't apply to default compose
    // addressing headers.  Or tags headerName.
    if (aCol.id == "formatCol" &&
        ((this.headerType == TM.v.HEADER[TM.v.HEADER.COMPOSE] &&
          TM.v.composeHeadersDefault.has(headerName)) ||
        ((this.headerType != TM.v.HEADER[TM.v.HEADER.COMPOSE] &&
          headerName == "tags"))))
      return false;

    let [headerCategory, headerValue] = this.getHeaderValue(headerName);
    if (!headerValue || !headerValue[TM.v.kMapped])
      return false;

    if (aCol.id == "formatCol" &&
        this.headerType != TM.v.HEADER[TM.v.HEADER.COMPOSE])
      this.openFormatPopup(aRow, aCol);

    return true;
  },

  setCellText: function(aRow, aCol, aText) {
TM.u.log.debug("TM.Preferences.setCellText: row:text - "+aRow+":"+aText);
    // Here only if in editor, and in a column that is editable.
    let headerName = this.visibleArray[aRow][this.kName];
    let [headerCategory, headerValue] = this.getHeaderValue(headerName);

    if (aCol.id == "nameCol") {
      // New custom header. Remove placeholder row.
      this.deleteHeader(headerName);
      if (aText)
        // If entered a name, lowercase it and add it.
        this.addHeader(aText.toLowerCase());
    }

    if (aCol.id == "displayNameCol" &&
        aText != this.getCellText(aRow, aCol)) {
      // The display name for headers.  Compose is strict rfc.
      if (this.headerType == TM.v.HEADER[TM.v.HEADER.COMPOSE]) {
        if (this.hasValidChar(aText))
          headerValue[TM.v.kDisplayName] = aText;
        else
          this.treeBox.invalidateRow(aRow);
      }
      else {
        let text = aText.replace(/␠/g, " ");
        headerValue[TM.v.kDisplayName] = text || null;
      }
    }

    if (aCol.id == "formatCol" &&
        this.headerType == TM.v.HEADER[TM.v.HEADER.COMPOSE])
      headerValue[TM.v.kFormat] = aText.replace(/␠/g, " ");

TM.u.log.debug("TM.Preferences.setCellText: new headerCategory:headerName:headerValue - "+
               headerCategory+":"+headerName+":"+headerValue);

    this.isNameColEditable = false;
    this.headersTree.focus();
  },

  hasNextSibling: function(aRow, aAfterIndex) {
    let thisLevel = this.getLevel(aRow);
    for (let t = aAfterIndex + 1; t < this.visibleArray.length; t++) {
      let nextLevel = this.getLevel(t);
      if (nextLevel == thisLevel)
        return true;
      if (nextLevel < thisLevel)
        break;
    }
    return false;
  },

  toggleOpenState: function(aRow) {
    let item = this.visibleArray[aRow];
    if (!item[this.kIsContainer])
      return;

    if (item[this.kIsOpen]) {
      item[this.kIsOpen] = false;

      let thisLevel = this.getLevel(aRow);
      let deletecount = 0;
      for (let t = aRow + 1; t < this.visibleArray.length; t++) {
        if (this.getLevel(t) > thisLevel)
          deletecount++;
        else
          break;
      }
      if (deletecount) {
        this.visibleArray.splice(aRow + 1, deletecount);
        this.treeBox.rowCountChanged(aRow + 1, -deletecount);
      }
    }
    else {
      item[this.kIsOpen] = true;

      let headerCategory = this.visibleArray[aRow][this.kCategoryName];
      let headers = [...this[headerCategory].keys()];
TM.u.log.debug("TM.Preferences.toggleOpenState: "+headerCategory+":"+headers);
      for (let i = 0; i < headers.length; i++)
        this.visibleArray.splice(aRow + i + 1, 0, [headers[i], false]);
      this.treeBox.rowCountChanged(aRow + 1, headers.length);
    }
    this.treeBox.invalidateRow(aRow);
    this.setOpenState();
  }

};

window.addEventListener("load", TM.Preferences.onLoad, true);
